/* $Id: offline.c,v 1.4 2000/01/02 12:04:58 linus Exp $
 *
 * Copyright (C) 1999  Lysator Computer Club,
 *                     Linkoping University, Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author: Linus Tolke
 */

#include <config.h>

#include <sys/param.h>

#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <kom-errno.h>

#include <misc-parser.h>

#include "xoutput.h"
#include "copy.h"
#include "error.h"
#include "cache.h"
#include "quit.h"
#include "main.h"

#define Export

/*
 * This file contains commands for offline reading.
 * Offline reading is a reduced form of reading when it is only possible to:
 * - List conferences with unread.
 * - Go to a conference that is downloaded.
 * - Read unread articles.
 * - Reread articles that is already downloaded.
 * - Mark an article.
 * - Comment articles.
 * - Write a new article in a downloaded conference.
 * - Write a letter to a downloaded person.
 *
 * Normally we work like this:
 *  1. Connect to the remote server (dialup...)
 *  2. Start the client and login.
 *  3. Issue a download command and wait for completion.
 *  4. Exit the client.
 *  5. Disconnect from the remote server (hangup...)
 *  6. Enter the client in offline-reading mode.
 *  7. Read and comment.
 *  8. Exit the client.
 *  9. Connect to the remote server (dialup...)
 * 10. Start the client and login.
 * 11. Issue an upload command and wait for completion.
 * 12. Exit the client.
 * 13. Disconnect from the remote server (hangup...)
 */

/* 
 * The offline mechanism requires a database that saves articles, 
 * persons and commands to disk for later processing.
 * The files of this database are:
 * $HOME/.lyskom-tty-offline-HOST:PORT/confstat-NR
 * $HOME/.lyskom-tty-offline-HOST:PORT/map-NR (includes timestamp).
 * $HOME/.lyskom-tty-offline-HOST:PORT/membership
 * $HOME/.lyskom-tty-offline-HOST:PORT/text-NR
 * $HOME/.lyskom-tty-offline-HOST:PORT/textstat-NR
 * $HOME/.lyskom-tty-offline-HOST:PORT/command-NR
 *
 * Alternativ r /tmp/<cuserid>-lyskom-tty-offline-HOST:PORT/.
 *
 * Commands are:
 * - mark as read
 * - create text (in conference, comment to, footnote to...)
 * - mark/unmark
 *
 * The database is implemented by the store_* and fetch_* functions here.
 */
static
void cache_file_name(char * path, const char * kind, long nr)
{
    static char basic_path[MAXPATHLEN];
    static int basic_path_tested = 0;

    if (!basic_path_tested)
    {
	char * home;
	int alternative;

	for (alternative = 1; alternative < 3; alternative++)
	{
	    switch (alternative)
	    {
	    case 1:
		home = getenv("HOME");
		if (home != NULL)
		{
		    sprintf(basic_path, "%s/.lyskom-tty-offline-%s:%d", 
			    home, server_name, portno);
		    break;
		}
		continue;

	    case 2:
		sprintf(basic_path, "/tmp/%s-lyskom-tty-offline-%s:%d",
#ifdef HAVE_CUSERID
			cuserid(NULL),
#else
			getlogin(),
#endif
			server_name, portno);
		break;
	    }

	    if (mkdir(basic_path, 0700) == 0
		|| errno == EEXIST)
	    {
		/* Ok! */
		/* Lets create a file and see if that works. */
		FILE * testfp = NULL;

		sprintf(path, "%sTESTING", basic_path);
		if ((testfp = fopen(path, "w")) == NULL)
		    continue;

		fclose(testfp);
		unlink(path);

		basic_path_tested = 1;
		break;
	    }
	}
	
	if (!basic_path_tested)
	{
	    /* We could not find any working alternative. */
	    fprintf(stderr, "Cannot find or create a working cache directory.\n");
	    abort();
	}
    }

    sprintf(path, "%s/%s-%d", basic_path, kind, nr);
}


FILE *
open_cache_file(const char * kind, long nr, const char * rw)
{
    char path[MAXPATHLEN];

    cache_file_name(path, kind, nr);

    return fopen(path, rw);
}

/* Routines to write different objects. */
void
fprintf_Time(FILE * fp, Time * t)
{
    fprintf(fp, "%lu %lu %lu %lu %lu %lu %lu %lu %lu",
	    t->tm_sec,
	    t->tm_min,
	    t->tm_hour,
	    t->tm_mday,
	    t->tm_mon,
	    t->tm_year,
	    t->tm_wday,
	    t->tm_yday,
	    t->tm_isdst);
}

void
fprintf_Membership(FILE * fp, Membership * ms)
{
    fprintf_Time(fp, &ms->last_time_read);
    fprintf(fp, " %lu %lu ", ms->conf_no, ms->priority);
    fprintf(fp, "%lu %lu ", ms->last_text_read, ms->no_of_read);
    if (ms->no_of_read > 0)
    {
	int i;

	fprintf(fp, "{ ");
	for (i = 0; i < ms->no_of_read; i++)
	{
	    fprintf(fp, "%lu ", ms->read_texts[i]);
	}
	fprintf(fp, "}");
    }
    else
	fprintf(fp, "*");
}

void
fprintf_Membership_list(FILE * fp, Membership_list * ml)
{
    int i;

    fprintf(fp, "%lu { ", ml->no_of_confs);
    for (i = 0; i < ml->no_of_confs; i++)
    {
	fprintf_Membership(fp, &ml->confs[i]);
	fprintf(fp, " ");
    }
    fprintf(fp, "}");
}

void
fprintf_String(FILE * fp, String * s)
{
    fprintf(fp, "%luH", s->len);
    s_fputs(fp, *s);
}

void
fprintf_Conf_type(FILE * fp, Conf_type * ct)
{
    fprintf(fp, "%1lu%1lu%1lu%1lu",
	    ct->rd_prot,
	    ct->original,
	    ct->secret,
	    ct->letter_box);
}

void
fprintf_Conference(FILE * fp, Conference * cs)
{
    fprintf_String(fp, &cs->name);
    fprintf(fp, " ");
    fprintf_Conf_type(fp, &cs->type);
    fprintf(fp, " ");
    fprintf_Time(fp, &cs->creation_time);
    fprintf(fp, " ");
    fprintf_Time(fp, &cs->last_written);
    fprintf(fp, " %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",
	    cs->creator,
	    cs->presentation,
	    cs->supervisor,
	    cs->permitted_submitters,
	    cs->super_conf,
	    cs->msg_of_day,
	    cs->nice,
	    cs->no_of_members,
	    cs->first_local_no,
	    cs->no_of_texts);
}


void
fprintf_Text_list(FILE * fp, Text_list * tl)
{
    fprintf(fp, "%lu %lu ", tl->first_local_no, tl->no_of_texts);
    if (tl->no_of_texts > 0)
    {
	int i;

	fprintf(fp, "{ ");
	for (i = 0; i < tl->no_of_texts; i++)
	    fprintf(fp, "%lu ", tl->texts[i]);
	fprintf(fp, "}");
    }
    else
	fprintf(fp, "*");
}



/* Routines to store and fetch. */
int
store_person_number(Pers_no pers)
{
    FILE * fp;

    if (trace > 1)
	xprintf("store_person_number\n");

    if ((fp = open_cache_file("person", 0, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf(fp, "%lu", pers);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

int
fetch_person_number(Pers_no * pers)
{
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_person_number\n");
    
    if ((fp = open_cache_file("person", 0, "r")) == NULL)
    {
	return FAILURE;
    }

    *pers = parse_long(fp);
    fclose(fp);
    return *pers != 0;
}


int
store_Membership_list(Membership_list * ml)
{
    FILE * fp;

    if (trace > 1)
	xprintf("store_Membership_list\n");

    if ((fp = open_cache_file("membership", 0, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf_Membership_list(fp, ml);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

int
fetch_Membership_list(Membership_list * ml)
{
    int res;
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_Membership_list\n");
    
    if ((fp = open_cache_file("membership", 0, "r")) == NULL)
    {
	return FAILURE;
    }

    res = parse_membership_list(fp, ml);
    fclose(fp);
    return res;
}

int
store_Conference(Conf_no cno, Conference * cs)
{
    FILE * fp;
    if (trace > 1)
	xprintf("store_Conference %ld\n", cno);

    if ((fp = open_cache_file("confstat", cno, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf_Conference(fp, cs);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

int
fetch_Conference(Conf_no cno, Conference * cs)
{
    int res;
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_Conference %ld\n", cno);

    if ((fp = open_cache_file("confstat", cno, "r")) == NULL)
    {
	return FAILURE;
    }

    res = parse_conference(fp, cs);
    fclose(fp);
    return res;
}

int
store_map(Conf_no cno, Text_list * tl, Time * ts)
{
    FILE * fp;
    if (trace > 1)
	xprintf("store_map %ld\n", cno);

    if ((fp = open_cache_file("map", cno, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf_Text_list(fp, tl);
    fprintf(fp, "\n");
    fprintf_Time(fp, ts);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}
int
fetch_map(Conf_no cno, Text_list * tl, Time * ts)
{
    int res;
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_map %ld\n", cno);

    if ((fp = open_cache_file("map", cno, "r")) == NULL)
    {
	return FAILURE;
    }

    res = parse_text_list(fp, tl);
    if (res == FAILURE)
    {
	fclose(fp);
	return FAILURE;
    }
    res = parse_time(fp, ts);
    fclose(fp);
    return res;
}

int
store_text(Text_no tno, String * tm)
{
    FILE * fp;
    if (trace > 1)
	xprintf("store_text %ld\n", tno);

    if ((fp = open_cache_file("text", tno, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf_String(fp, tm);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

int
fetch_text(Text_no tno, String * tm)
{
    int res;
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_text %ld\n", tno);

    if ((fp = open_cache_file("text", tno, "r")) == NULL)
    {
	return FAILURE;
    }

    res = parse_string(fp, tm);
    fclose(fp);
    return res;
}

int
store_Text_stat(Text_no tno, Text_stat * ts)
{
    FILE * fp;
    if (trace > 1)
	xprintf("store_Text_stat %ld\n", tno);

    if ((fp = open_cache_file("textstat", tno, "w")) == NULL)
    {
	return FAILURE;
    }

    fprintf_Time(fp, &ts->creation_time);
    fprintf(fp, " %lu %lu %lu %lu %lu ",
	    ts->author,
	    ts->no_of_lines,
	    ts->no_of_chars,
	    ts->no_of_marks,
	    ts->no_of_misc);

    if (ts->no_of_misc > 0)
    {
        int i;

	fprintf(fp, "{ ");

	for (i = 0; i < ts->no_of_misc; i++)
	{
	    switch (ts->misc_items[i].type)
	    {
	    case recpt:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.recipient);
		break;
	    case cc_recpt:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.cc_recipient);
		break;
	    case loc_no:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.local_no);
		break;
	    case rec_time:
		fprintf(fp, "%lu ", ts->misc_items[i].type);
		fprintf_Time(fp, &ts->misc_items[i].datum.received_at);
		break;
	    case comm_to:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.comment_to);
		break;
	    case comm_in:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.commented_in);
		break;
	    case footn_to:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.footnote_to);
		break;
	    case footn_in:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.footnoted_in);
		break;
	    case sent_by:
		fprintf(fp, "%lu %lu ", 
			ts->misc_items[i].type, 
			ts->misc_items[i].datum.sender);
		break;
	    case sent_at:
		fprintf(fp, "%lu ", ts->misc_items[i].type);
		fprintf_Time(fp, &ts->misc_items[i].datum.sent_at);
		break;
	    default:
		fprintf(stderr, "Unknown text-stat type: %lu\n", 
			ts->misc_items[i].type);
		break;
	    }
	}
	fprintf(fp, "}");
    }
    else
	fprintf(fp, "*");
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

int
fetch_Text_stat(Text_no tno, Text_stat * ts)
{
    int res;
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_Text_stat %ld\n", tno);

    if ((fp = open_cache_file("textstat", tno, "r")) == NULL)
    {
	return FAILURE;
    }

    res = parse_text_stat(fp, ts);
    fclose(fp);
    return res;
}

/* ================================================================
 * Here are functions that look exactly like the access functions
 * above. The difference is that they are to keep track of what we
 * have read.
 * ================================================================
 */
/*
 * store_mark_in_conf appends to the file.
 */
int
store_mark_in_conf(Conf_no cno, Local_text_no ltno)
{
    FILE * fp;

    if (trace > 1)
	xprintf("store_mark_in_conf\n");

    if ((fp = open_cache_file("marks-in-conf", cno, "a")) == NULL)
    {
	return FAILURE;
    }

    fprintf(fp, "%lu", ltno);
    fprintf(fp, "\n");
    fclose(fp);
    return OK;
}

/*
 * fetch_marks_in_conf
 * Returns a list of all texts to be marked as read in this conference.
 */
int
fetch_marks_in_conf(Conf_no cno, int * len, Local_text_no * ltnos)
{
    FILE * fp;

    if (trace > 1)
	xprintf("fetch_marks_in_conf\n");
    
    if ((fp = open_cache_file("marks-in-conf", cno, "r")) == NULL)
    {
	return FAILURE;
    }

    /* Ej klar */
    /* *pers = parse_long(fp); */
    fclose(fp);
    return 17;
}

/*
 * done_with_marks_in_conf
 */
int
done_with_marks_in_conf(Conf_no cno)
{
    char path[MAXPATHLEN];

    cache_file_name(path, "marks-in-conf", cno);

    if (unlink(path) < 0)
    {
        return FAILURE;
    }

    return OK;
}


/* ================================================================
 * Functions that "replace" complex functions.
 *
 * ================================================================
 */
static int		memb_list_read = 0;
static Membership_list	memb_list = EMPTY_MEMBERSHIP_LIST_i;
static int		get_unread_confs_pos;

/*
 * offline_get_unread_confs_start
 *
 * Initiate the reading of the membership.
 */
int
offline_get_unread_confs_start()
{
  /*+++ Inte bra! Mste lsa saved_membership_list frn cache.c */ 
    if (!memb_list_read)
    {
	if (fetch_Membership_list(&memb_list) != OK)
	{
	    return 0;
	}
    }

    get_unread_confs_pos = -1;
    return 1;
}

/*
 * offline_get_unread_confs_next
 *
 * Returns the conf no of the next conference with unread news.
 * If there is none, return 0.
 */
Conf_no
offline_get_unread_confs_next()
{
    while (++get_unread_confs_pos < memb_list.no_of_confs)
    {
        Text_list 	ltl = EMPTY_TEXT_LIST_i;
	Time		lt = NULL_TIME_i;
	Conference	conf = EMPTY_CONFERENCE_i;
	Conf_no		confno = memb_list.confs[get_unread_confs_pos].conf_no;

	if (memb_list.confs[get_unread_confs_pos].no_of_read > 0)
	{
	    if (fetch_map(confno, &ltl, &lt) == FAILURE)
	    {
	        /* Disqualify conferences that we havn't fetched the map for.
		 */
	        continue;
	    }
	    else
	    {
		release_text_list(&ltl);
	    }

	    return memb_list.confs[get_unread_confs_pos].conf_no;
	}

	if (fetch_Conference(confno, &conf) == OK)
	{
	    if (memb_list.confs[get_unread_confs_pos].last_text_read < 
		conf.first_local_no + conf.no_of_texts - 1)
	    {
		release_conf(&conf);

		if (fetch_map(confno, &ltl, &lt) == FAILURE)
		{
		    /* Disqualify conferences that we havn't fetched
		     * the map for.
		     */
		    continue;
		}
		else
		{
		    release_text_list(&ltl);
		}

		return memb_list.confs[get_unread_confs_pos].conf_no;
	    }
	    else
	    {
		release_conf(&conf);
	    }
	}
    }

    return (Conf_no)0;
}

/*
 * offline_get_unread_confs_count
 *
 * Calculates how many unread confs we have.
 * The algorithm is:
 * Read the membership, for each of the conferences read the conf-stat and
 * if that conf has unread add one.
 * Since texts that are marked as read by this or earlier sessions of 
 * offline reading shall not count, then we have to read the map for that
 * conference and see if any of those texts are read.
 * This we will leave out since the test is complicated and I have to write
 * the read marking code first. +++
 * Another solution is to update the membership as we read.
 */
int
offline_get_unread_confs_count()
{
    int			count = 0;

    if (!offline_get_unread_confs_start())
    {
	return 0;
    }

    while (offline_get_unread_confs_next())
    {
	count++;
    }
    return count;
}


/*
 * offline_get_unread_confs
 *
 * Returns a list of conferences with unread.
 */
Success
offline_get_unread_confs(Conf_no_list * list)
{
#define FUNCTION "offline_get_unread_confs()"

    Conf_no cno;

    release_conf_no_list(list);

    offline_get_unread_confs_start();

    list->no_of_confs = 0;
    /* This will always be enough. */
    list->conf_nos =
        (Conf_no*)zmalloc(memb_list.no_of_confs * sizeof(Conf_no));
    if (list->conf_nos == NULL)
        fatal2(CLIENT_OUT_OF_MEMORY, "zmalloc(%d) failed.",
	       memb_list.no_of_confs * sizeof(Conf_no));

    while ((cno = offline_get_unread_confs_next()) != (Conf_no)0)
    {
        list->conf_nos[list->no_of_confs++] = cno;
    }

    return OK;
#undef FUNCTION
}
